-- Al final de tot teniu les tecles per usar-lo
-- Un cop implementat pintor.hs, heu de compilar amb 
-- ghc --make visualPintor.hs pintor.hs
-- i executar amb 
-- ./visualPintor

import MRobot
import qualified Data.Maybe as DM
import qualified Data.Map as DMap
import qualified Control.Monad as CM
import qualified Graphics.UI.SDL as SDL
import qualified Graphics.UI.SDL.Image as SDLi

-- amb import qualified Data.Maybe as DM podem usar tot el que es defineix
-- al modul Data.Maybe pero cal posar DM. al davant

-- creem un a llista constant amb els grafics dels tipus de terreny del mapa

mapFilePaths:: [String]
mapFilePaths =["grafics/white_square.png",
               "grafics/blue_square.png",
               "grafics/yellow_square.png",
               "grafics/green_square.png",
               "grafics/red_square.png"]

blackFilePath :: String
blackFilePath = "grafics/black_square.png"

cartutxFilePaths:: [String]
cartutxFilePaths =["grafics/cartutx_white.png",
               "grafics/cartutx_blue.png",
               "grafics/cartutx_yellow.png",
               "grafics/cartutx_green.png",
               "grafics/cartutx_red.png"]

-- creem un a llista constant amb els grafics dels robots

-- definim una enumeració per les diferents formes del robot al mapa
data RobotPlaces = WBlackRobot | WhiteRobot | BlueRobot | YelowRobot | GreenRobot | RedRobot
     deriving (Bounded, Eq, Enum, Ord, Show)

obtenerRobotColor :: Robot -> RobotPlaces
obtenerRobotColor r 
		  | not (esPinta r) && not (esPintaDis r) = WBlackRobot 
		  | esBlanc (colorPinta r) = WhiteRobot
		  | esBlau (colorPinta r) = BlueRobot
		  | esGroc (colorPinta r) = YelowRobot
		  | esVerd (colorPinta r) = GreenRobot
		  | esVermell (colorPinta r) = RedRobot

robotFilePaths :: [String]
robotFilePaths = ["grafics/robot_white_black.png",
               	  "grafics/robot_white.png",
               	  "grafics/robot_blue.png",
               	  "grafics/robot_yellow.png",
               	  "grafics/robot_green.png",
               	  "grafics/robot_red.png"]

-- llista amb els tipus de colors
colorTypes = [getBlanc,getBlau,getGroc,getVerd,getVermell]

-- llista amb els tipus de robots (enumerats)
robotPlacesTypes :: [RobotPlaces]
robotPlacesTypes = enumFrom WBlackRobot

type Punt = (Int, Int)

data Cel = Col Color | Cart Color | Res
     deriving (Eq)

-- definim el mapa amb un Map de Punts a Colors
type TerrainMap = DMap.Map Punt Cel

obtenirCel :: Mapa -> Int -> Int -> Cel
obtenirCel m x y 
  | not (teColor m x y) && not (teCartutx m x y) = Res
  | teCartutx m x y = Cart (obtenirColorCartutx m x y)
  | otherwise = Col (obtenirColor m x y)

-- crea un mapa del terreny a partir del nostre mapa i les dimensions
-- fa un fols que crea totes les files usant makeRow
makeMap :: Int -> Int -> Mapa -> TerrainMap
makeMap h w mapa = foldl (\m x -> makeRow w x (obtenirCel mapa) m) DMap.empty [1..h]
  where
     makeRow :: Int -> Int -> (Int -> Int -> Cel) -> TerrainMap -> TerrainMap
     makeRow w x fvmap tileMap = 
         let tp = (zip [1..w] (map (\y->fvmap (x-1) (y-1)) [1..w])) in
         (foldr (\(y,t) m -> DMap.insert (x,y) t m)  tileMap tp)
-- els -1 son perque el nostre mapa comença en (0,0) i el del grafic va be que comenci en (1,1)

-- relaciona els tipus de terrenys amb superficies SDL
type ColorSurfaces = [(Color, SDL.Surface)]

-- aquesta funció carrega les imatges d'una llista i les relaciona amb els tipus de terrenys
loadSurfaces :: [String] -> IO ColorSurfaces
loadSurfaces paths = do
     tileSurfs <- mapM SDLi.load paths
     return $ zip colorTypes tileSurfs

-- relaciona els tipus de robots amb superficies SDL
type RobotSurfaces = [(RobotPlaces, SDL.Surface)]

-- aquesta funció carrega les imatges d'una llista i les relaciona amb els tipus de robot
loadRobotSurfaces :: [String] -> IO RobotSurfaces
loadRobotSurfaces paths = do
     tileSurfs <- mapM SDLi.load paths
     return $ zip robotPlacesTypes tileSurfs

getSurfCel :: ColorSurfaces -> SDL.Surface -> ColorSurfaces -> Cel -> SDL.Surface
getSurfCel terrainSurfs blackSurf cartutxSurfs Res = blackSurf
getSurfCel terrainSurfs blackSurf cartutxSurfs (Col c) = DM.fromJust $ lookup c terrainSurfs
getSurfCel terrainSurfs blackSurf cartutxSurfs (Cart c) = DM.fromJust $ lookup c cartutxSurfs

dibuixaMosaic :: SDL.Surface -> ColorSurfaces -> SDL.Surface -> ColorSurfaces -> TerrainMap -> Punt -> IO ()
dibuixaMosaic mainSurf terrainSurfs blackSurf cartutxSurfs tm (x,y) = do
     let sr = Just (SDL.Rect 0 0 64 64)
     let dr = Just $ getHexmapOffset 64 64 x y
     let tt = DM.fromJust $ DMap.lookup (x,y) tm
     let terrainSurf = getSurfCel terrainSurfs blackSurf cartutxSurfs tt
     SDL.blitSurface terrainSurf sr mainSurf dr
     return ()
-- mainSurf es la superficie total del mapa grafic
-- terrainSurf es la superficie (figura) que hem associat a tt
-- sr hi ha les dimensions del rectangle 64x64 
-- dr hi ha les dimensions del rectangle 64x64
-- blitSurface posa el rectangle sr tret de la superficie terrainSurf 
-- sobre la superficie mainSurf a la posició indicada pel rectangle dr

getHexmapOffset :: Int -> Int -> Int -> Int -> SDL.Rect
getHexmapOffset tileW tileH x y =
--     SDL.Rect adjX adjY 0 0
     SDL.Rect adjX adjY tileW tileH
  where
     adjX = (tileW * (x-1))
     adjY = (tileH * (y-1))
-- donada l'amplada i l'alçada de cada quadrat i una posició (considerant 
-- (1,1) com a origen) calcula la posició en la superficie total del grafic 
-- del mapa
	       

placeColor :: Mapa -> Int -> Int -> SDL.Surface -> ColorSurfaces -> SDL.Surface ->IO ()
placeColor m x y mainSurf colorSurfs blackSurf
 | not (teColor m x y) = do
     let sr = Just (SDL.Rect 0 0 64 64)
     let dr = Just $ getHexmapOffset 64 64 (x+1) (y+1)
     SDL.blitSurface blackSurf sr mainSurf dr
     return ()
 | otherwise = do
     let c = obtenirColor m x y
     let sr = Just (SDL.Rect 0 0 64 64)
     let dr = Just $ getHexmapOffset 64 64 (x+1) (y+1)
     let colorSurf = DM.fromJust $ lookup c colorSurfs
     SDL.blitSurface colorSurf sr mainSurf dr
     return ()
-- posa un quadrat d'un cert tipus de terreny en una posicio del la 
-- superficie del mapa (i rep la llista de superficies de terreny) 

placeRobot :: Robot ->SDL.Surface->RobotSurfaces->TerrainMap->IO ()
placeRobot r mainSurf robotSurfs tm = do
  let sr = Just (SDL.Rect 0 0 64 64)
  let dr = Just $ getHexmapOffset 64 64 (x+1) (y+1)
  let robotSurf = DM.fromJust $ lookup rc robotSurfs
  SDL.blitSurface robotSurf sr mainSurf dr
  return ()
  where (x,y) = obtenirPos r
        rc = obtenerRobotColor r

-- com l'anterior pero per robots, en aquest cas decideix el tipus de 
-- robot a usar segons el mapa del terreny.

main :: IO ()
main = do
     SDL.init [SDL.InitEverything]
     SDL.setVideoMode (64*wideMap) (64*highMap) 32 []
     SDL.setCaption "Robot Test" "video test"
     mainSurf <- SDL.getVideoSurface
     tileSurfs <- loadSurfaces mapFilePaths
     blackSurf <- SDLi.load blackFilePath
     tileCartutxSurf <- loadSurfaces cartutxFilePaths
     tileRobSurfs <- loadRobotSurfaces robotFilePaths
     mapM_ (dibuixaMosaic mainSurf tileSurfs blackSurf tileCartutxSurf robotMap) $ DMap.keys robotMap
     placeRobot robotInici mainSurf tileRobSurfs robotMap
     SDL.flip mainSurf
     eventLoop robotInici mapaIni mainSurf tileSurfs blackSurf tileRobSurfs
     mapM_ freeSurf tileSurfs 
     mapM_ freeSurf tileRobSurfs 
     SDL.quit
     -- print "done"
 where
     wideMap = ampladaMapa mapaIni
     highMap = alturaMapa mapaIni
     robotMap = makeMap wideMap highMap mapaIni
     freeSurf (_ , s) = SDL.freeSurface s
     eventLoop r m mainSurf tileSurfs blackSurf tileRobSurfs= SDL.waitEventBlocking 
                >>= (checkEvent r m mainSurf tileSurfs blackSurf tileRobSurfs)
     checkEvent _ _ _ _ _ _ (SDL.KeyUp (SDL.Keysym SDL.SDLK_ESCAPE _ _)) = return()
     checkEvent r m mainSurf tileSurfs blackSurf tileRobSurfs (SDL.KeyUp (SDL.Keysym SDL.SDLK_DOWN _ _)) =  do 
     		placeColor m x y mainSurf tileSurfs blackSurf
     		placeRobot r1 mainSurf tileRobSurfs robotMap
		SDL.flip mainSurf
	        eventLoop r1 m1 mainSurf tileSurfs blackSurf tileRobSurfs 
         where (x,y) = obtenirPos r
               (r1,m1) = (avanca (gDown r) m)
     checkEvent r m mainSurf tileSurfs blackSurf tileRobSurfs (SDL.KeyUp (SDL.Keysym SDL.SDLK_UP _ _)) =  do 
     		placeColor m x y mainSurf tileSurfs blackSurf
     		placeRobot r1 mainSurf tileRobSurfs robotMap
		SDL.flip mainSurf
	        eventLoop r1 m1 mainSurf tileSurfs blackSurf tileRobSurfs 
         where (x,y) = obtenirPos r
	       c = obtenirColor m x y
               (r1,m1) = (avanca (gUp r) m)
     checkEvent r m mainSurf tileSurfs blackSurf tileRobSurfs (SDL.KeyUp (SDL.Keysym SDL.SDLK_RIGHT _ _)) =  do
     		placeColor m x y mainSurf tileSurfs blackSurf
     		placeRobot r1 mainSurf tileRobSurfs robotMap
		SDL.flip mainSurf
	        eventLoop r1 m1 mainSurf tileSurfs blackSurf tileRobSurfs 
         where (x,y) = obtenirPos r
	       c = obtenirColor m x y
               (r1,m1) = (avanca (gRight r) m)
     checkEvent r m mainSurf tileSurfs blackSurf tileRobSurfs (SDL.KeyUp (SDL.Keysym SDL.SDLK_LEFT _ _)) =  do
     		placeColor m x y mainSurf tileSurfs blackSurf
     		placeRobot r1 mainSurf tileRobSurfs robotMap
		SDL.flip mainSurf
	        eventLoop r1 m1 mainSurf tileSurfs blackSurf tileRobSurfs 
         where (x,y) = obtenirPos r
	       c = obtenirColor m x y
               (r1,m1) = (avanca (gLeft r) m)
     checkEvent r m mainSurf tileSurfs blackSurf tileRobSurfs (SDL.KeyUp (SDL.Keysym SDL.SDLK_SPACE _ _)) =  do
     		placeRobot r1 mainSurf tileRobSurfs robotMap
		SDL.flip mainSurf
	        eventLoop r1 m mainSurf tileSurfs blackSurf tileRobSurfs 
         where (x,y) = obtenirPos r
               r1 = if (esPinta r) then (desactivarPintar r) else (activarPintar r)
     checkEvent r m mainSurf tileSurfs blackSurf tileRobSurfs (SDL.KeyUp (SDL.Keysym SDL.SDLK_DELETE _ _)) =  do
     		placeRobot r1 mainSurf tileRobSurfs robotMap
		SDL.flip mainSurf
	        eventLoop r1 m mainSurf tileSurfs blackSurf tileRobSurfs 
         where (x,y) = obtenirPos r
               r1 = if (esPintaDis r) then (desactivarPintarDis r) else (activarPintarDis r)
     checkEvent r m mainSurf tileSurfs blackSurf tileRobSurfs (SDL.KeyUp (SDL.Keysym SDL.SDLK_w _ _)) =  do
     		placeRobot r1 mainSurf tileRobSurfs robotMap
		SDL.flip mainSurf
	        eventLoop r1 m mainSurf tileSurfs blackSurf tileRobSurfs 
         where (x,y) = obtenirPos r
               r1 = canviaBlanc r
     checkEvent r m mainSurf tileSurfs blackSurf tileRobSurfs (SDL.KeyUp (SDL.Keysym SDL.SDLK_b _ _)) =  do
     		placeRobot r1 mainSurf tileRobSurfs robotMap
		SDL.flip mainSurf
	        eventLoop r1 m mainSurf tileSurfs blackSurf tileRobSurfs 
         where (x,y) = obtenirPos r
               r1 = canviaBlau r
     checkEvent r m mainSurf tileSurfs blackSurf tileRobSurfs (SDL.KeyUp (SDL.Keysym SDL.SDLK_y _ _)) =  do
     		placeRobot r1 mainSurf tileRobSurfs robotMap
		SDL.flip mainSurf
	        eventLoop r1 m mainSurf tileSurfs blackSurf tileRobSurfs 
         where (x,y) = obtenirPos r
               r1 = canviaGroc r
     checkEvent r m mainSurf tileSurfs blackSurf tileRobSurfs (SDL.KeyUp (SDL.Keysym SDL.SDLK_g _ _)) =  do
     		placeRobot r1 mainSurf tileRobSurfs robotMap
		SDL.flip mainSurf
	        eventLoop r1 m mainSurf tileSurfs blackSurf tileRobSurfs 
         where (x,y) = obtenirPos r
               r1 = canviaVerd r
     checkEvent r m mainSurf tileSurfs blackSurf tileRobSurfs (SDL.KeyUp (SDL.Keysym SDL.SDLK_r _ _)) =  do
     		placeRobot r1 mainSurf tileRobSurfs robotMap
		SDL.flip mainSurf
	        eventLoop r1 m mainSurf tileSurfs blackSurf tileRobSurfs 
         where (x,y) = obtenirPos r
               r1 = canviaVermell r
     checkEvent r m mainSurf tileSurfs blackSurf tileRobSurfs _   = 
            eventLoop r m mainSurf tileSurfs blackSurf tileRobSurfs 

-- Per usar-lo:
--
-- espai activa i desactiva pintar (si activa és en blanc)
-- supr activa i desactiva pintar discontinu (si activa és en blanc)
--      per activar una accio cal que l'altre no estigui activa
-- w activa el color blanc (si pintar o pintar disc actiu)
-- b activa el color blau (si pintar o pintar disc actiu)
-- y activa el color groc (si pintar o pintar disc actiu)
-- g activa el color verd (si pintar o pintar disc actiu)
-- r activa el color vermell (si pintar o pintar disc actiu)
-- cursor amunt per moure's una posició en direcció amunt
-- cursor avall per moure's una posició en direcció avall
-- cursor dreta per moure's una posició en direcció dreta
-- cursor esquerra per moure's una posició en direcció esquerra
-- ESC per sortir
